from escape.core import *
from escape.scattering import *
from escape.version import __version__


def require(version):
    '''Require can be used to check the minimum version required to run a ESCAPE
    application. For example, you can start your application code like this::

        import escape
        escape.require('1.0.1')

    If a user attempts to run your application with a version of ESCAPE that is
    older than the specified version, an Exception is raised.

    The ESCAPE version string is built like this::

        X.Y.Z[-tag[-tagrevision]]

        X is the major version
        Y is the minor version
        Z is the bugfixes revision

    The tag is optional, but may be one of 'dev', 'alpha', or 'beta'.
    The tagrevision is the revision of the tag.

    .. warning::

        You must not ask for a version with a tag, except -dev. Asking for a
        'dev' version will just warn the user if the current ESCAPE
        version is not a -dev, but it will never raise an exception.
        You must not ask for a version with a tagrevision.

    '''

    def parse_version(version):
        # check for tag
        tag = None
        tagrev = None
        if '-' in version:
            v = version.split('-')
            if len(v) == 2:
                version, tag = v
            elif len(v) == 3:
                version, tag, tagrev = v
            else:
                raise Exception('Revision format must be X.Y.Z[-tag]')

        # check x y z
        v = version.split('.')
        if len(v) != 3:
            if 'dev0' in v:
                tag = v.pop()
            else:
                raise Exception('Revision format must be X.Y.Z[-tag]')
        return [int(x) for x in v], tag, tagrev

    # user version
    revision, tag, tagrev = parse_version(version)
    # current version
    sysrevision, systag, _ = parse_version(__version__)

    # ensure that the required version don't contain tag, except dev
    if tag not in (None, 'dev'):
        raise Exception('Revision format must not have any tag except "dev"')
    if tag == 'dev' and systag != 'dev':
        print('Application requested a -dev version of ESCAPE. '
                       '(You have %s, but the application requires %s)' % (
                           __version__, version))
    # not tag rev (-alpha-1, -beta-x) allowed.
    if tagrev is not None:
        raise Exception('Revision format must not contain any tagrevision')

    # finally, checking revision
    if sysrevision < revision:
        raise Exception('The version of ESCAPE installed on this system '
                        'is too old. '
                        '(You have %s, but the application requires %s)' % (
                            __version__, version))
        
