from .module import Module
from .mlp import MLP
from .regressor import Regressor, loads_torch_regressor


from escape.core.setting import int_setting_obj
from escape.core.model import modelstack


def torch_regressor(name, mstack, *, module=MLP, nsamples=1000, ntests=100,
              data_noisifiers=None,
              coordinates_noisifiers=None,
              data_normalizers=None, **kwargs):
    """
    Creates torch regressor object. 
    
    :Parameters:
        `name`: string
            Name of regressor object.
        `mstack`: modelstack_obj, list of models or model_obj
            Models to be optimized.
        `module`: Python class or instance of `Module`
            Name of module class or instance of `Module` to be used.
        `nsamples`: positive integer
            Number of training datasets to be generated.
        `ntests`: positive integer
            Number of test datasets to be generated.
        `data_noisifiers`: callable or list of callables
            Noisifier class instance to be used to noisify generated training data.
        `coordinates_noisifiers`: callable or list of callables
            Instance to be used to noisify generated training data coordinates.
        `data_normalizers`: callable or list of callables
            Method or instance to be used to normalize generated training data.
        `kwargs`: dict
            Keyword arguments for regressor.
            
    :return:
        instance of regressor.torch.Regressor
        
    """
    nsamples = int_setting_obj.convert(nsamples)
    ntests = int_setting_obj.convert(ntests)
    stack = modelstack(f"Modelstack: {name}", mstack)
    resobj = Regressor(name, stack, module=module, 
        nsamples=nsamples, ntests=ntests, 
        data_noisifiers=data_noisifiers,
        coordinates_noisifiers=coordinates_noisifiers,
        data_normalizers=data_normalizers, **kwargs)
    return resobj