import time, io
import threading 
import numpy as np

import ipywidgets.widgets as widgets

from escape.utils.widgets.base import EscapeObjectLayout
from escape.utils.widgets.modelstack import MSObjectLayout
from escape.utils.widgets.plot import Plot2dWidget
from ipywidgets.widgets.widget_box import VBox


class RegressorObjectLayout(MSObjectLayout):
    
    def __init__(self, escobj, **kwargs):
        modelstack = escobj.modelstack

        self.regressor = escobj
        self.kwargs=kwargs
        self.modelstack = modelstack
        MSObjectLayout.__init__(self, modelstack, parameter_checkbox_property="trained", **kwargs)
        self.update()
    
    def init_header(self, **kwargs):

        def on_train_click(btn):
            try:
                self.trainbtn.disabled = True
                self.resetbtn.disabled = True
                self.shakebtn.disabled = True
                self.nextbtn.disabled = True
                self.prevbtn.disabled = True
                self.disable_parameters(True)
                
                escobj = self.regressor
                escobj.on_train_begin_method = self.on_train_begin
                escobj.on_train_end_method = self.on_train_end
                escobj.on_epoch_begin_method = self.on_epoch_begin
                escobj.on_epoch_end_method = self.on_epoch_end

                score=self.regressor.train(True)
                self.status_label.value = '<b>Score:</b> {0:.4g}'.format(score)
                self.update_parameter_sliders(silent=True)
                self._update(buf=False)
            except Exception as err:
                self.status_label.value = '<b>Error:</b> {0}'.format(err)
            finally:
                self.trainbtn.disabled = False
                self.resetbtn.disabled = False
                self.shakebtn.disabled = False
                self.nextbtn.disabled = False
                self.prevbtn.disabled = False
                self.disable_parameters(False)
                
        def on_next_click(btn):
            try:
                self.regressor.next_test()
                self.update_parameter_sliders(silent=True)
                self._update(buf=False)
            except Exception as err:
                print(err)
                
        def on_prev_click(btn):
            try:
                self.regressor.prev_test()
                self.update_parameter_sliders(silent=True)
                self._update(buf=False)
            except Exception as err:
                print(err)
            
        def on_reset_click(btn):
            try:
                self.regressor.reset()
                self.update_parameter_sliders(silent=True)
                self._update(buf=False)
            except Exception as err:
                print(err)
            
        def on_shake_click(btn):
            try:
                self.modelstack.shake()
                self.update_parameter_sliders(silent=True)
                self._update(buf=False)
            except Exception as err:
                print(err)
        
        btnlay = widgets.Layout( width='40px' )
            
        self.trainbtn = widgets.Button(
                description='',
                disabled=False,
                tooltip='Train regressor and predict',
                icon='train',
                layout=btnlay)
        self.trainbtn.on_click(on_train_click)
        
        self.nextbtn = widgets.Button(
                description='',
                disabled=False,
                tooltip='Shows solution for the next test data',
                icon='arrow-right',
                layout=btnlay)
        self.nextbtn.on_click(on_next_click)
        
        self.prevbtn = widgets.Button(
                description='',
                disabled=False,
                tooltip='Shows solution for the prev test data',
                icon='arrow-left',
                layout=btnlay)
        self.prevbtn.on_click(on_prev_click)
        
        self.resetbtn = widgets.Button(
                description='',
                disabled=False,
                tooltip='Reset parameters values',
                icon='refresh',
                layout=btnlay)
        
        self.resetbtn.on_click(on_reset_click)
        
        self.shakebtn = widgets.Button(
                description='',
                disabled=False,
                tooltip='Shake parameters values',
                icon='random',
                layout=btnlay)
        
        self.shakebtn.on_click(on_shake_click)
                
        toolbar = widgets.HBox(children=[
            self.trainbtn,
            self.prevbtn,
            self.nextbtn,
            self.resetbtn,
            self.shakebtn,
            ])
        
        return toolbar
    
    def init_center(self, **kwargs):
        plots = super(RegressorObjectLayout, self).init_center(**kwargs)
        return plots

    def init_footer(self, **kwargs):
        epochs_plot = Plot2dWidget(self, width=850, height=180, xlabel="Epochs",
                                 ylabel="Error", tight=True)
        status_label = widgets.HTML('Status: ')
        self.status_label = status_label
        self.epochs_plot = epochs_plot
        return widgets.VBox([status_label, epochs_plot])

    def on_train_begin(self, net, X=None, y=None, **kwargs):
        self.status_label.value = '<b>Training...</b>'

    def on_train_end(self, net, X=None, y=None, **kwargs):
        self.status_label.value = '<b>Done</b>'
        escobj = self.regressor
        escobj.on_train_begin_method = None
        escobj.on_train_end_method = None
        escobj.on_epoch_begin_method = None
        escobj.on_epoch_end_method = None

    def on_epoch_begin(self, net, dataset_train=None, dataset_valid=None, **kwargs):
        self.status_label.value = '<b>Epoch {0}</b>'.format(self.regressor.trainer.history[-1, 'epoch'])

    def on_epoch_end(self, net, dataset_train=None, dataset_valid=None, **kwargs):
        try:
            history = self.regressor.trainer.history
            epochs = np.asarray(history[:, 'epoch'])
            trl = np.asarray(history[:, 'train_loss'])
            epochs_plot = self.epochs_plot
            epochs_plot.update(epochs[1:], trl)
        except Exception as err:
            self.status_label.value = '<b>Error:</b> {0}'.format(err)
            
    